@@ -49,6 +49,9 @@
 
 #define N_LISTENERS 6
 
+#define VINO_DBUS_BUS_NAME  "org.gnome.Vino"
+#define VINO_DBUS_INTERFACE "org.gnome.VinoScreen"
+
 typedef struct {
   GladeXML    *xml;
   GConfClient *client;
@@ -75,6 +78,9 @@ typedef struct {
   guint        use_password : 1;
 } VinoPreferencesDialog;
 
+static void
+vino_preferences_dialog_update_url_label (VinoPreferencesDialog *dialog);
+
 static char *
 vino_preferences_dialog_get_password_from_keyring (VinoPreferencesDialog *dialog)
 {
@@ -96,10 +102,9 @@ vino_preferences_dialog_get_password_from_keyring (VinoPreferencesDialog *dialog
 		5900,           /* port     */
 		&matches);
 
-  if (result != GNOME_KEYRING_RESULT_OK)
+  if (result != GNOME_KEYRING_RESULT_OK || matches == NULL || matches->data == NULL)
     return NULL;
 
-  g_assert (matches != NULL && matches->data != NULL);
 
   found_item = (GnomeKeyringNetworkPasswordData *) matches->data;
 
@@ -455,12 +460,19 @@ vino_preferences_vnc_password_notify (GConfClient           *client,
   if (!entry->value || entry->value->type != GCONF_VALUE_STRING)
     return;
 
+  password = NULL;
+
   password_b64 = gconf_value_get_string (entry->value);
 
-  blob_len = 0;
-  blob = g_base64_decode (password_b64, &blob_len);
+  if (password_b64 && *password_b64)
+    {
+      blob_len = 0;
+      blob = g_base64_decode (password_b64, &blob_len);
+    
+      password = g_strndup ((char *) blob, blob_len);
 
-  password = g_strndup ((char *) blob, blob_len);
+      g_free (blob);
+    }
 
   if (!password || !password [0])
     {
@@ -479,7 +491,6 @@ vino_preferences_vnc_password_notify (GConfClient           *client,
     }
 
   g_free (password);
-  g_free (blob);
 }
 
 static void
@@ -497,7 +508,7 @@ vino_preferences_dialog_password_changed (GtkEntry              *entry,
     {
       gconf_client_unset (dialog->client, VINO_PREFS_VNC_PASSWORD, NULL);
     }
-  else
+  else if (strlen (password) > 1)
     {
       char *password_b64;
 
@@ -531,13 +542,16 @@ vino_preferences_dialog_setup_password_entry (VinoPreferencesDialog *dialog)
 
       password_b64 = gconf_client_get_string (dialog->client, VINO_PREFS_VNC_PASSWORD, NULL);
 
-      blob_len = 0;
-      blob = g_base64_decode (password_b64, &blob_len);
+      if (password_b64 && *password_b64)
+        {
+           blob_len = 0;
+           blob = g_base64_decode (password_b64, &blob_len);
 
-      password = g_strndup ((char *) blob, blob_len);
-
-      g_free (blob);
-      g_free (password_b64);
+           password = g_strndup ((char *) blob, blob_len);
+        
+           g_free (blob);
+           g_free (password_b64);
+        }
 
       password_in_keyring = FALSE;
     }
@@ -595,12 +609,48 @@ vino_preferences_dialog_setup_icons (VinoPreferencesDialog *dialog)
                                 GTK_ICON_SIZE_DIALOG);
 }
 
+static void
+vino_preferences_server_updated (DBusGProxy *proxy,
+                                 const char *name,
+                                 const char *prev_owner,
+                                 const char *new_owner,
+                                 gpointer user_data)
+{
+  if ( (new_owner) && (!strcmp (name, VINO_DBUS_BUS_NAME)) )
+    vino_preferences_dialog_update_url_label ( (VinoPreferencesDialog *) user_data);
+}
+
+static void
+vino_preferences_start_listening (VinoPreferencesDialog *dialog)
+{
+  DBusGConnection *connection;
+  GError          *error;
+  DBusGProxy      *proxy;
+  
+  error = NULL;
+  connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
+  if (!connection)
+    {
+      g_printerr (_("Failed to open connection to bus: %s\n"),
+                  error->message);
+      g_error_free (error);
+      return;
+    }
+
+  proxy = dbus_g_proxy_new_for_name (connection,
+                                     DBUS_SERVICE_DBUS,
+                                     DBUS_PATH_DBUS,
+                                     DBUS_INTERFACE_DBUS);
+
+  dbus_g_proxy_add_signal (proxy, "NameOwnerChanged",
+                           G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INVALID);
+  dbus_g_proxy_connect_signal (proxy, "NameOwnerChanged",
+                          G_CALLBACK (vino_preferences_server_updated), dialog, NULL);
+}
+
 static int
 vino_preferences_get_server_port (VinoPreferencesDialog *dialog)
 {
-#define VINO_DBUS_BUS_NAME  "org.gnome.Vino"
-#define VINO_DBUS_INTERFACE "org.gnome.VinoScreen"
-
 #define VINO_DEFAULT_PORT   5900
 #define VINO_MIN_PORT       5000
 #define VINO_MAX_PORT       6000
@@ -639,8 +689,7 @@ vino_preferences_get_server_port (VinoPreferencesDialog *dialog)
                           G_TYPE_INT, &port,
                           G_TYPE_INVALID))
     {
-      g_printerr ("Failed to call remote GetServerPort function: %s\n",
-                  error->message);
+      g_object_unref (proxy);
       g_error_free (error);
       dbus_g_connection_unref (connection);
       return 0;
@@ -654,13 +703,64 @@ vino_preferences_get_server_port (VinoPreferencesDialog *dialog)
   else
     return port;
 
-#undef VINO_DBUS_BUS_NAME
-#undef VINO_DBUS_INTERFACE
 #undef VINO_DEFAULT_PORT
 #undef VINO_MIN_PORT
 #undef VINO_MAX_PORT
 }
 
+#ifdef VINO_ENABLE_HTTP_SERVER
+static int
+vino_preferences_get_http_server_port (VinoPreferencesDialog *dialog)
+{
+#define VINO_HTTP_DEFAULT_PORT   5800
+  DBusGConnection *connection;
+  GError          *error;
+  DBusGProxy      *proxy;
+  int              port;
+  GdkScreen       *screen;
+  char            *obj_path;
+  
+  error = NULL;
+  connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
+  if (!connection)
+    {
+      g_printerr (_("Failed to open connection to bus: %s\n"),
+                  error->message);
+      g_error_free (error);
+      return 0;
+    }
+
+  screen = gtk_window_get_screen (GTK_WINDOW (dialog->dialog));
+
+  obj_path = g_strdup_printf ("/org/gnome/vino/screens/%d",
+                              gdk_screen_get_number (screen));
+
+  proxy = dbus_g_proxy_new_for_name (connection,
+                                     VINO_DBUS_BUS_NAME,
+                                     obj_path,
+                                     VINO_DBUS_INTERFACE);
+
+  g_free (obj_path);
+
+  if (!dbus_g_proxy_call (proxy, "GetHttpServerPort", &error,
+                          G_TYPE_INVALID,
+                          G_TYPE_INT, &port,
+                          G_TYPE_INVALID))
+    {
+      g_object_unref (proxy);
+      g_error_free (error);
+      dbus_g_connection_unref (connection);
+      return VINO_HTTP_DEFAULT_PORT;
+    }
+
+  g_object_unref (proxy);
+  dbus_g_connection_unref (connection);
+
+  return port;
+#undef VINO_HTTP_DEFAULT_PORT
+}
+#endif /* VINO_ENABLE_HTTP_SERVER */
+
 static char *
 vino_preferences_get_local_hostname (void)
 {
@@ -695,14 +795,18 @@ vino_preferences_dialog_get_server_command (VinoPreferencesDialog *dialog)
   int   server_port;
 
   local_host  = vino_preferences_get_local_hostname ();
-  server_port = vino_preferences_get_server_port (dialog);
-
   if (!local_host)
     {
-      return g_strdup_printf ("vncviewer localhost:%d", server_port);
+      local_host = g_strdup_printf ("localhost");
     }
 
+#ifdef VINO_ENABLE_HTTP_SERVER
+  server_port = vino_preferences_get_http_server_port (dialog);
+  server_url = g_strdup_printf ("http://%s:%d", local_host, server_port);
+#else
+  server_port = vino_preferences_get_server_port (dialog);
   server_url = g_strdup_printf ("vncviewer %s:%d", local_host, server_port);
+#endif
 
   g_free (local_host);
 
@@ -1003,6 +1107,7 @@ main (int argc, char **argv)
       return 1;
     }
 
+  vino_preferences_start_listening (&dialog);
   gtk_main ();
 
   vino_preferences_dialog_finalize (&dialog);
