@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2006 Jonh Wendell
+ * Copyright (C) 2007 Mark McLoughlin
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -18,6 +19,7 @@
  *
  * Authors:
  *      Jonh Wendell <wendell@bani.com.br>
+ *      Mark McLoughlin <mark@skynet.ie>
  */
 
 #include <config.h>
@@ -28,11 +30,20 @@
 #include <gtk/gtk.h>
 #include <string.h>
 
+#ifdef VINO_ENABLE_LIBNOTIFY
+#include <libnotify/notify.h>
+#endif
+
 struct _VinoStatusIconPrivate
 {
   GtkMenu    *menu;
   VinoServer *server;
   GSList     *clients;
+  GtkWidget  *disconnect_dialog;
+
+#ifdef VINO_ENABLE_LIBNOTIFY
+  NotifyNotification *new_client_notification;
+#endif
 };
 
 G_DEFINE_TYPE (VinoStatusIcon, vino_status_icon, GTK_TYPE_STATUS_ICON);
@@ -43,11 +54,25 @@ enum
   PROP_SERVER
 };
 
+typedef struct
+{
+  VinoStatusIcon *icon;
+  VinoClient     *client;
+}VinoStatusIconNotify;
+
+static gboolean vino_status_icon_show_new_client_notification (gpointer user_data);
+
 static void
 vino_status_icon_finalize (GObject *object)
 {
   VinoStatusIcon *icon = VINO_STATUS_ICON (object);
 
+#ifdef VINO_ENABLE_LIBNOTIFY
+  if (icon->priv->new_client_notification)
+    g_object_unref (icon->priv->new_client_notification);
+  icon->priv->new_client_notification = NULL;
+#endif
+
   if (icon->priv->menu)
     gtk_widget_destroy (GTK_WIDGET(icon->priv->menu));
   icon->priv->menu = NULL;
@@ -56,6 +81,10 @@ vino_status_icon_finalize (GObject *object)
     g_slist_free (icon->priv->clients);
   icon->priv->clients = NULL;
 
+  if (icon->priv->disconnect_dialog)
+    gtk_widget_destroy (icon->priv->disconnect_dialog);
+  icon->priv->disconnect_dialog = NULL;
+
   G_OBJECT_CLASS (vino_status_icon_parent_class)->finalize (object);
 }
 
@@ -179,8 +208,8 @@ static void
 vino_status_icon_help (VinoStatusIcon *icon)
 {
   vino_status_icon_spawn_command (icon,
-                                  "yelp ghelp:user-guide?goscustdesk-90",
-                                  _("There was an error displaying jelp:\n %s"));
+                                  "gnome-open ghelp:user-guide?goscustdesk-90",
+                                  _("There was an error displaying help:\n %s"));
 }
 
 static void
@@ -234,13 +263,57 @@ vino_status_icon_about (VinoStatusIcon *icon)
                          NULL);
 }
 
-static gboolean
-vino_status_icon_disconnect_confirm (VinoClient *client)
+static void
+vino_status_icon_disconnect_client (VinoStatusIconNotify *a, gint response)
+{
+
+  GSList *l;
+  GSList *next;
+
+  VinoStatusIcon *icon    = a->icon;
+  VinoClient     *client  = a->client;
+
+  gtk_widget_destroy (icon->priv->disconnect_dialog);
+  icon->priv->disconnect_dialog = NULL;
+
+  if (response != GTK_RESPONSE_OK)
+  {
+    g_free (a);
+    return;
+  }
+
+  if (client)
+  {
+    if (g_slist_find (icon->priv->clients, client))
+      vino_client_disconnect (client);
+  }
+  else
+    for (l = icon->priv->clients; l; l = next)
+      {
+        VinoClient *client = l->data;
+
+        next = l->next;
+
+        vino_client_disconnect (client);
+      }
+
+  g_free (a);
+}
+
+static void
+vino_status_icon_disconnect_confirm (VinoStatusIconNotify *a)
 {
-  GtkWidget *dialog;
   char      *primary_msg;
   char      *secondary_msg;
-  gboolean   retval;
+
+  VinoStatusIcon *icon    = a->icon;
+  VinoClient     *client  = a->client;
+
+  if (icon->priv->disconnect_dialog)
+  {
+    gtk_window_present (GTK_WINDOW(icon->priv->disconnect_dialog));
+    return;
+  }
 
   if (client != NULL)
     {
@@ -255,53 +328,26 @@ vino_status_icon_disconnect_confirm (VinoClient *client)
       secondary_msg = g_strdup (_("All remote users will be disconnected. Are you sure?"));
     }
 
-  dialog = gtk_message_dialog_new (NULL,
-                                   GTK_DIALOG_DESTROY_WITH_PARENT,
-                                   GTK_MESSAGE_QUESTION,
-                                   GTK_BUTTONS_CANCEL,
-                                   "%s",
-                                   primary_msg);
+  icon->priv->disconnect_dialog = gtk_message_dialog_new (NULL,
+                                                          GTK_DIALOG_DESTROY_WITH_PARENT,
+                                                          GTK_MESSAGE_QUESTION,
+                                                          GTK_BUTTONS_CANCEL,
+                                                          "%s",
+                                                          primary_msg);
 
-  gtk_window_set_skip_taskbar_hint (GTK_WINDOW (dialog), FALSE);
+  gtk_window_set_skip_taskbar_hint (GTK_WINDOW (icon->priv->disconnect_dialog), FALSE);
 
-  gtk_dialog_add_button (GTK_DIALOG (dialog), _("Disconnect"), GTK_RESPONSE_OK);
+  gtk_dialog_add_button (GTK_DIALOG (icon->priv->disconnect_dialog), _("Disconnect"), GTK_RESPONSE_OK);
 
-  gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog), "%s", secondary_msg);
+  gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (icon->priv->disconnect_dialog),
+                                            "%s", secondary_msg);
 
-  retval = gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_OK;
-
-  gtk_widget_destroy (dialog);
+  g_signal_connect_swapped (icon->priv->disconnect_dialog, "response",
+                            G_CALLBACK (vino_status_icon_disconnect_client), (gpointer) a);  
+  gtk_widget_show_all (GTK_WIDGET(icon->priv->disconnect_dialog));
 
   g_free (primary_msg);
   g_free (secondary_msg);
-
-  return retval;
-}
-
-static void
-vino_status_icon_disconnect_client (VinoClient *client)
-{
-  if (vino_status_icon_disconnect_confirm (client))
-    vino_client_disconnect (client);
-}
-
-static void
-vino_status_icon_disconnect_all_clients (VinoStatusIcon *icon)
-{
-  GSList *l;
-  GSList *next;
-
-  if (!vino_status_icon_disconnect_confirm (NULL))
-    return;
-
-  for (l = icon->priv->clients; l; l = next)
-    {
-      VinoClient *client = l->data;
-
-      next = l->next;
-
-      vino_client_disconnect (client);
-    }
 }
 
 static void
@@ -312,6 +358,7 @@ vino_status_icon_popup_menu (GtkStatusIcon *status_icon,
   VinoStatusIcon *icon = VINO_STATUS_ICON (status_icon);
   GtkWidget      *item;
   GSList         *l;
+  VinoStatusIconNotify *a;
 
   icon->priv->menu = (GtkMenu*) gtk_menu_new ();
 
@@ -330,8 +377,13 @@ vino_status_icon_popup_menu (GtkStatusIcon *status_icon,
   item  = gtk_image_menu_item_new_with_label (_("Disconnect all"));
   gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item),
                                  gtk_image_new_from_stock (GTK_STOCK_NETWORK, GTK_ICON_SIZE_MENU));
+
+  a = g_new (VinoStatusIconNotify, 1);
+  a->icon   = icon;
+  a->client = NULL;
+
   g_signal_connect_swapped (item, "activate",
-                            G_CALLBACK (vino_status_icon_disconnect_all_clients), icon);
+                            G_CALLBACK (vino_status_icon_disconnect_confirm), (gpointer) a);
   gtk_widget_show (item);
   gtk_menu_shell_append (GTK_MENU_SHELL (icon->priv->menu), item);
 
@@ -339,6 +391,10 @@ vino_status_icon_popup_menu (GtkStatusIcon *status_icon,
     {
       VinoClient *client = l->data;
       char       *str;
+      
+      a = g_new (VinoStatusIconNotify, 1);
+      a->icon   = icon;
+      a->client = client;
 
       str = g_strdup_printf (_("Disconnect %s"), vino_client_get_hostname (client));
 
@@ -346,7 +402,7 @@ vino_status_icon_popup_menu (GtkStatusIcon *status_icon,
       gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item),
                                      gtk_image_new_from_stock (GTK_STOCK_NETWORK, GTK_ICON_SIZE_MENU));
       g_signal_connect_swapped (item, "activate",
-                                G_CALLBACK (vino_status_icon_disconnect_client), client);
+                                G_CALLBACK (vino_status_icon_disconnect_confirm), (gpointer) a);
       gtk_widget_show (item);
       gtk_menu_shell_append (GTK_MENU_SHELL (icon->priv->menu), item);
       
@@ -420,6 +476,14 @@ vino_status_icon_add_client (VinoStatusIcon *icon,
   icon->priv->clients = g_slist_append (icon->priv->clients, client);
 
   vino_status_icon_update_tooltip (icon);
+
+  VinoStatusIconNotify *a;
+  a = g_new (VinoStatusIconNotify, 1);
+  a->icon   = icon;
+  a->client = client;
+  g_timeout_add (1000, 
+                 vino_status_icon_show_new_client_notification,
+                 (gpointer) a);
 }
 
 gboolean
@@ -463,3 +527,90 @@ vino_status_icon_class_init (VinoStatusIconClass *klass)
 
   g_type_class_add_private (gobject_class, sizeof (VinoStatusIconPrivate));
 }
+
+#ifdef VINO_ENABLE_LIBNOTIFY
+static void
+vino_status_handle_new_client_notification_closed (VinoStatusIcon *icon)
+{
+  g_object_unref (icon->priv->new_client_notification);
+  icon->priv->new_client_notification = NULL;
+}
+#endif /* VINO_ENABLE_LIBNOTIFY */
+
+static gboolean
+vino_status_icon_show_new_client_notification (gpointer user_data)
+{
+#ifdef VINO_ENABLE_LIBNOTIFY
+#define NOTIFICATION_TIMEOUT 5
+
+  GError     *error;
+  const char *summary;
+  char       *body;
+
+  VinoStatusIconNotify *a = (VinoStatusIconNotify *)user_data;
+  VinoStatusIcon *icon    = a->icon;
+  VinoClient     *client  = a->client;
+
+  if (vino_server_get_prompt_enabled (icon->priv->server))
+  {
+    g_free (user_data);
+    return FALSE;
+  }
+
+  if (!notify_is_initted () &&  !notify_init (_("GNOME Remote Desktop")))
+    {
+      g_printerr (_("Error initializing libnotify\n"));
+      g_free (user_data);
+      return FALSE;
+    }
+
+  if (icon->priv->new_client_notification)
+    {
+      notify_notification_close (icon->priv->new_client_notification, NULL);
+      g_object_unref (icon->priv->new_client_notification);
+      icon->priv->new_client_notification = NULL;
+    }
+
+  if (vino_server_get_view_only (icon->priv->server))
+    {
+      summary = _("Another user is viewing your desktop");
+      body = g_strdup_printf (_("A user on the computer '%s' is remotely viewing your desktop."),
+                              vino_client_get_hostname (client));
+    }
+  else
+    {
+      summary = _("Another user is controlling your desktop");
+      body = g_strdup_printf (_("A user on the computer '%s' is remotely controlling your desktop."),
+                              vino_client_get_hostname (client));
+    }
+
+  icon->priv->new_client_notification =
+    notify_notification_new_with_status_icon (summary,
+                                              body,
+                                              "gnome-remote-desktop",
+                                              GTK_STATUS_ICON (icon));
+
+  g_free (body);
+
+  g_signal_connect_swapped (icon->priv->new_client_notification, "closed",
+                            G_CALLBACK (vino_status_handle_new_client_notification_closed),
+                            icon);
+
+  notify_notification_set_timeout (icon->priv->new_client_notification,
+                                   NOTIFICATION_TIMEOUT * 1000);
+
+  error = NULL;
+  if (!notify_notification_show (icon->priv->new_client_notification, &error))
+    {
+      g_printerr (_("Error while displaying notification bubble: %s\n"),
+                  error->message);
+      g_error_free (error);
+    }
+
+  g_free (user_data);
+
+#undef NOTIFICATION_TIMEOUT
+#endif /* VINO_ENABLE_LIBNOTIFY */
+
+  return FALSE;
+}
